@@ -1,8 +1,8 @@
 /*
  * The MIT License (MIT)
  *
- * Copyright (c) 2015-2019 Derick Rethans
- * Copyright (c) 2018 MongoDB, Inc.
+ * Copyright (c) 2015-2022 Derick Rethans
+ * Copyright (c) 2018,2021 MongoDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -30,6 +30,10 @@
 # include "timelib_config.h"
 #endif
 
+#define TIMELIB_VERSION 202117
+#define TIMELIB_EXTENDED_VERSION 20211701
+#define TIMELIB_ASCII_VERSION "2021.17"
+
 #include <stdlib.h>
 #include <stdbool.h>
 #include <limits.h>
@@ -144,6 +148,44 @@ typedef struct _tlocinfo
 	char *comments;
 } tlocinfo;
 
+#define TIMELIB_POSIX_TRANS_TYPE_JULIAN_NO_FEB29   1
+#define TIMELIB_POSIX_TRANS_TYPE_JULIAN_FEB29      2
+#define TIMELIB_POSIX_TRANS_TYPE_MWD               3
+
+typedef struct _timelib_posix_trans_info
+{
+	int type; // 1=Jn, 2=n, 3=Mm.w.d
+	union {
+		int days;
+		struct {
+			int month;
+			int week;
+			int dow;
+		} mwd;
+	};
+	int hour;
+} timelib_posix_trans_info;
+
+typedef struct _timelib_posix_str
+{
+	char        *std;
+	timelib_sll  std_offset;
+	char        *dst;
+	timelib_sll  dst_offset;
+
+	timelib_posix_trans_info *dst_begin;
+	timelib_posix_trans_info *dst_end;
+
+	int type_index_std_type;  // index into tz->type
+	int type_index_dst_type;  // index into tz->type
+} timelib_posix_str;
+
+typedef struct _timelib_posix_transitions {
+	size_t      count;
+	timelib_sll times[6];
+	timelib_sll types[6];
+} timelib_posix_transitions;
+
 typedef struct _timelib_tzinfo
 {
 	char    *name;
@@ -173,6 +215,9 @@ typedef struct _timelib_tzinfo
 	tlinfo  *leap_times;
 	unsigned char bc;
 	tlocinfo location;
+
+	char              *posix_string;
+	timelib_posix_str *posix_info;
 } timelib_tzinfo;
 
 typedef struct _timelib_rel_time {
@@ -277,6 +322,7 @@ typedef struct _timelib_abbr_info {
 #define TIMELIB_ERR_FORMAT_LITERAL_MISMATCH    0x224
 #define TIMELIB_ERR_MIX_ISO_WITH_NATURAL       0x225
 
+#define TIMELIB_ZONETYPE_NONE   0
 #define TIMELIB_ZONETYPE_OFFSET 1
 #define TIMELIB_ZONETYPE_ABBR   2
 #define TIMELIB_ZONETYPE_ID     3
@@ -296,10 +342,10 @@ typedef struct _timelib_error_container {
 } timelib_error_container;
 
 typedef struct _timelib_tz_lookup_table {
-	char       *name;
+	const char *name;
 	int         type;
 	float       gmtoffset;
-	char       *full_tz_name;
+	const char *full_tz_name;
 } timelib_tz_lookup_table;
 
 typedef struct _timelib_tzdb_index_entry {
@@ -308,7 +354,7 @@ typedef struct _timelib_tzdb_index_entry {
 } timelib_tzdb_index_entry;
 
 typedef struct _timelib_tzdb {
-	char                           *version;
+	const char                     *version;
 	int                             index_size;
 	const timelib_tzdb_index_entry *index;
 	const unsigned char            *data;
@@ -320,12 +366,14 @@ typedef struct _timelib_tzdb {
 # define timelib_calloc  calloc
 # define timelib_strdup  strdup
 # define timelib_free    free
+# if defined(WIN32) || defined(_WIN32) || defined(__WIN32) && !defined(__CYGWIN__)
+#  define TIMELIB_USE_BUILTIN_STRNDUP 1
+# else
+#  define TIMELIB_USE_BUILTIN_STRNDUP 0
+#  define timelib_strndup strndup
+# endif
 #endif
 
-#define TIMELIB_VERSION 202002
-#define TIMELIB_EXTENDED_VERSION 20202001
-#define TIMELIB_ASCII_VERSION "2020.02"
-
 #define TIMELIB_NONE             0x00
 #define TIMELIB_OVERRIDE_TIME    0x01
 #define TIMELIB_NO_CLONE         0x02
@@ -333,7 +381,8 @@ typedef struct _timelib_tzdb {
 #define TIMELIB_UNSET   -99999
 
 /* An entry for each of these error codes is also in the
- * timelib_error_messages array in timelib.c */
+ * timelib_error_messages array in timelib.c.
+ * Codes 0x00, 0x07, and 0x09 are warnings only. */
 #define TIMELIB_ERROR_NO_ERROR                            0x00
 #define TIMELIB_ERROR_CANNOT_ALLOCATE                     0x01
 #define TIMELIB_ERROR_CORRUPT_TRANSITIONS_DONT_INCREASE   0x02
@@ -341,6 +390,9 @@ typedef struct _timelib_tzdb {
 #define TIMELIB_ERROR_CORRUPT_NO_ABBREVIATION             0x04
 #define TIMELIB_ERROR_UNSUPPORTED_VERSION                 0x05
 #define TIMELIB_ERROR_NO_SUCH_TIMEZONE                    0x06
+#define TIMELIB_ERROR_SLIM_FILE                           0x07 /* Warns if the file is SLIM, but we can't read it */
+#define TIMELIB_ERROR_CORRUPT_POSIX_STRING                0x08
+#define TIMELIB_ERROR_EMPTY_POSIX_STRING                  0x09 /* Warns if the POSIX string is empty, but still produces results */
 
 #ifdef __cplusplus
 extern "C" {
@@ -531,7 +583,7 @@ void timelib_fill_holes(timelib_time *parsed, timelib_time *now, int options);
  *
  * The returned char* is not duplicated, and should not be freed.
  */
-char *timelib_timezone_id_from_abbr(const char *abbr, timelib_long gmtoffset, int isdst);
+const char *timelib_timezone_id_from_abbr(const char *abbr, timelib_long gmtoffset, int isdst);
 
 /* Returns an array of known time zone abbreviations
  *
@@ -588,6 +640,12 @@ void timelib_strtointerval(const char *s, size_t len,
  */
 void timelib_update_ts(timelib_time* time, timelib_tzinfo* tzi);
 
+/**
+ * Returns the number of days from the y/m/d fields of 'time' since the Unix
+ * Epoch.
+ */
+timelib_sll timelib_epoch_days_from_time(timelib_time *time);
+
 /**
  * Takes the information from the y/m/d/h/i/s fields and makes sure their
  * values are in the right range.
@@ -610,6 +668,14 @@ void timelib_do_rel_normalize(timelib_time *base, timelib_rel_time *rt);
 
 /* From unixtime2tm.c */
 
+/**
+ * Takes the unix timestamp in seconds from 'ts', and calculates y, m, and d,
+ * in the proleptic Gregorian calendar.
+ *
+ * It uses the algorithm from howardhinnant.github.io/date_algorithms.html
+ */
+void timelib_unixtime2date(timelib_sll ts, timelib_sll *y, timelib_sll *m, timelib_sll *d);
+
 /**
  * Takes the unix timestamp in seconds from 'ts' and populates the y/m/d/h/i/s
  * fields of 'tm' without taking time zones into account
@@ -669,14 +735,26 @@ int timelib_timezone_id_is_valid(const char *timezone, const timelib_tzdb *tzdb)
 
 /**
  * Converts the binary stored time zone information from 'tzdb' for the time
- * zone 'timeozne' into a structure the library can use for calculations.
+ * zone 'timezone' into a structure the library can use for calculations.
  *
  * The function can be used on both timelib_builtin_db as well as a time zone
  * db as opened by timelib_zoneinfo.
+ *
+ * 'error_code' must not be a null pointer, and will always be written to. If
+ * the value is TIMELIB_ERROR_NO_ERROR then the file was parsed without
+ * problems.
+ *
  * The function will return null upon failure, and also set an error code
- * through 'error_code'. 'error_code' must not be a null pointer. The error
- * code is one of the TIMELIB_ERROR_* constants as listed above. These error
- * constants can be converted into a string by timelib_get_error_message.
+ * through 'error_code'.
+ *
+ * The error code is one of the TIMELIB_ERROR_* constants as listed above.
+ * These error constants can be converted into a string by
+ * timelib_get_error_message.
+ *
+ * If the function returns not-null, the 'error_code' might have a non-null
+ * value that can be used to detect incompatibilities. The only one that is
+ * currently detected is whether the file is a 'slim' file, in which case
+ * 'error_code' will be set to TIMELIB_ERROR_SLIM_FILE.
  *
  * This function allocates memory for the new time zone structure, which must
  * be freed after use. Although it is recommended that a cache of each used
@@ -721,6 +799,16 @@ timelib_time_offset *timelib_get_time_zone_info(timelib_sll ts, timelib_tzinfo *
  */
 timelib_sll timelib_get_current_offset(timelib_time *t);
 
+/**
+ * Returns whether the timezone information in *one and *two are the same
+ *
+ * A timezone is considered the same if:
+ * - the ->zone_type values are the same for *one and *two
+ * - for TYPE_ABBR and TYPE_OFFSET, ->z + (->dst * 3600), is the same
+ * - for TYPE_ID, the zone's names are the same
+ */
+int timelib_same_timezone(timelib_time *one, timelib_time *two);
+
 /**
  * Displays debugging information about the time zone information in 'tz'.
  */
@@ -883,6 +971,17 @@ void timelib_decimal_hour_to_hms(double h, int *hour, int *min, int *sec);
  */
 void timelib_hms_to_decimal_hour(int hour, int min, int sec, double *h);
 
+/**
+ * Converts hour/min/sec/micro sec values into a decimal hour
+ */
+void timelib_hmsf_to_decimal_hour(int hour, int min, int sec, int us, double *h);
+
+/**
+ * Converts hour/min/sec values into seconds
+ */
+timelib_sll timelib_hms_to_seconds(timelib_sll h, timelib_sll m, timelib_sll s);
+
+
 /* from astro.c */
 
 /**
@@ -943,6 +1042,15 @@ int timelib_astro_rise_set_altitude(timelib_time *time, double lon, double lat,
  */
 timelib_rel_time *timelib_diff(timelib_time *one, timelib_time *two);
 
+/**
+ * Calculates the difference in full days between two times
+ *
+ * The result is the number of full days between 'one' and 'two'. It does take
+ * into account 23 and 25 hour (and variants) days when the zone_type
+ * is TIMELIB_ZONETYPE_ID and have the same TZID for 'one' and 'two'.
+ */
+int timelib_diff_days(timelib_time *one, timelib_time *two);
+
 /**
  * Adds the relative time information 'interval' to the base time 't'.
  *
@@ -950,6 +1058,7 @@ timelib_rel_time *timelib_diff(timelib_time *one, timelib_time *two);
  * complex statements such as "next workday".
  */
 timelib_time *timelib_add(timelib_time *t, timelib_rel_time *interval);
+timelib_time *timelib_add_wall(timelib_time *t, timelib_rel_time *interval);
 
 /**
  * Subtracts the relative time information 'interval' to the base time 't'.
@@ -959,6 +1068,19 @@ timelib_time *timelib_add(timelib_time *t, timelib_rel_time *interval);
  * workday".
  */
 timelib_time *timelib_sub(timelib_time *t, timelib_rel_time *interval);
+timelib_time *timelib_sub_wall(timelib_time *t, timelib_rel_time *interval);
+
+/* from parse_posix.c */
+
+void timelib_posix_str_dtor(timelib_posix_str *ps);
+
+timelib_posix_str* timelib_parse_posix_str(const char *posix);
+
+/**
+ * Calculate the two yearly to/from DST
+ */
+void timelib_get_transitions_for_year(timelib_tzinfo *tz, timelib_sll year, timelib_posix_transitions *transitions);
+
 
 #ifdef __cplusplus
 } /* extern "C" */
